"""
dlg_lowpass_components generators
"""
import logging

import numpy as np
from dlg import droputils
from dlg.common.reproducibility.constants import system_summary
from dlg.drop import BarrierAppDROP
from dlg.meta import (
    dlg_batch_input,
    dlg_batch_output,
    dlg_component,
    dlg_float_param,
    dlg_int_param,
    dlg_streaming_input,
    dlg_dict_param,
)

logger = logging.getLogger(__name__)


##
# @brief LP_SignalGenerator
# @details Generates a noisy sine signal for filtering. Effectively an input generator.
#
# @par EAGLE_START
# @param category PythonApp
# @param[in] cparam/appclass appclass/dlg_lowpass_components.LPSignalGenerator/String/readonly/
#     \~English Import direction for application class
# @param[in] cparam/execution_time Execution Time/5/Float/readonly/False//False/
#     \~English Estimated execution time
# @param[in] cparam/num_cpus No. of CPUs/1/Integer/readonly/False//False/
#     \~English Number of cores used
# @param[in] aparam/length Signal length/256/Integer/readwrite/
#     \~English Length of the output signal
# @param[in] aparam/samplerate Sample rate/5000/Integer/readwrite/
#     \~English The sample rate of the signal
# @param[in] aparam/frequencies Signal frequencies/{}/Json/readwrite/
#     \~English A dictionary containing a single list of values - the frequencies incorporated in the original signal.
# @param[in] aparam/noise_params Noise parameters/{}/Json/readwrite/
#     \~English A dictionary containing several values defining the properties of an interleaved noise. noise, stddiv-deviation, frequency, random randomseed, noisemultiplier
# @param[out] port/signal Signal/float/
#     \~English Numpy array containing final signal (purely real (floats))
# @par EAGLE_END
class LPSignalGenerator(BarrierAppDROP):
    """
    Generates a noisy sine signal for filtering. Effectively an input generator.
    """

    component_meta = dlg_component(
        "LPSignalGen",
        "Low-pass filter example signal generator",
        [None],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )

    # default values
    length = dlg_int_param("length", 256)
    samplerate = dlg_int_param("samplerate", 5000)
    frequencies = dlg_dict_param("frequencies", {"values": [440, 800, 1000, 2000]})
    noise = dlg_dict_param(
        "noise", {}
    )  # {'noise': 0.0, 'stddiv': 1.0, 'frequency': 666, 'randomseed': 42})
    series = None

    def initialize(self, **kwargs):
        super(LPSignalGenerator, self).initialize(**kwargs)

    def add_noise(
        self, series: np.array, mean, std, freq, sample_rate, seed, alpha=0.1
    ):
        """
        A noise to the provided signal by producing random values of a given frequency
        :param series: The input (and output) numpy array signal series
        :param mean: The average value
        :param std: The standard deviation of the value
        :param freq: The frequency of the noisy signal
        :param sample_rate: The sample rate of the input series
        :param seed: The random randomseed
        :param alpha: The multiplier
        :return: The input series with noisy values added
        """
        np.random.seed(seed)
        samples = alpha * np.random.normal(mean, std, size=len(series))
        for i in range(len(series)):
            samples[i] += np.sin(2 * np.pi * i * freq / sample_rate)
        np.add(series, samples, out=series)
        return series

    def gen_sig(self):
        """
        Generates an initial signal
        :return: Numpy array of signal values.
        """
        series = np.zeros(self.length, dtype=np.float64)
        for freq in self.frequencies["values"]:
            for i in range(self.length):
                series[i] += np.sin(2 * np.pi * i * freq / self.samplerate)
        return series

    def run(self):
        """
        Called by DALiuGE to start signal generation. Conditionally adds noise if parameters are set
        :return: Writes signal to output ports.
        """
        outs = self.outputs
        if len(outs) < 1:
            raise Exception("At least one output required for %r" % self)
        self.series = self.gen_sig()
        if len(self.noise) > 0:
            if "noisemultiplier" in self.noise:
                self.noise["noisemultiplier"] = 1 / self.noise["noisemultiplier"]
            self.series = self.add_noise(
                self.series,
                self.noise["noise"],
                self.noise["stddiv"],
                self.noise["frequency"],
                self.samplerate,
                self.noise["randomseed"],
                self.noise.get("noisemultiplier", 0.1),
            )

        data = self.series.tobytes()
        for output in outs:
            output.len = len(data)
            output.write(data)

    def generate_recompute_data(self):
        # This will do for now
        return {
            "length": self.length,
            "sample_rate": self.samplerate,
            "frequencies": self.frequencies,
            "status": self.status,
            "system": system_summary(),
        }


##
# @brief LP_WindowGenerator
# @details Generates a Hann window for low-pass filtering.
# @par EAGLE_START
# @param category PythonApp
# @param[in] cparam/appclass appclass/dlg_lowpass_components.LPWindowGenerator/String/readonly/
#     \~English Import direction for application class
# @param[in] cparam/execution_time Execution Time/5/Float/readonly/False//False/
#     \~English Estimated execution time
# @param[in] cparam/num_cpus No. of CPUs/1/Integer/readonly/False//False/
#     \~English Number of cores used
# @param[in] aparam/length Signal length/256/Integer/readwrite/
#     \~English Length of the output signal
# @param[in] aparam/samplerate Sample rate/5000/Integer/readwrite/
#     \~English The sample rate of the signal
# @param[in] aparam/cutoff Filter cutoff/600/Integer/readwrite/
#     \~English The frequency of the low-pass filter
# @param[out] port/window Window/float/
#     \~English Numpy array containing final signal (purely real (floats))
# @par EAGLE_END
class LPWindowGenerator(BarrierAppDROP):
    """
    Generates a Hann window for low-pass filtering.
    """

    component_meta = dlg_component(
        "LPWindowGen",
        "Low-pass filter example window generator",
        [None],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )

    # default values
    length = dlg_int_param("length", 256)
    cutoff = dlg_int_param("cutoff", 600)
    samplerate = dlg_int_param("samplerate", 5000)
    series = None

    def sinc(self, x_val: np.float64):
        """
        Computes the sin_c value for the input float
        :param x_val:
        """
        if np.isclose(x_val, 0.0):
            return 1.0
        return np.sin(np.pi * x_val) / (np.pi * x_val)

    def gen_win(self):
        """
        Generates the window values.
        :return: Numpy array of window series.
        """
        alpha = 2 * self.cutoff / self.samplerate
        win = np.zeros(self.length, dtype=np.float64)
        for i in range(int(self.length)):
            ham = 0.54 - 0.46 * np.cos(
                2 * np.pi * i / int(self.length)
            )  # Hamming coefficient
            hsupp = i - int(self.length) / 2
            win[i] = ham * alpha * self.sinc(alpha * hsupp)
        return win

    def run(self):
        """
        Called by DALiuGE to start drop execution
        :return:
        """
        outs = self.outputs
        if len(outs) < 1:
            raise Exception("At least one output required for %r" % self)
        self.series = self.gen_win()
        data = self.series.tobytes()
        for output in outs:
            output.len = len(data)
            output.write(data)

    def generate_recompute_data(self):
        output = dict()
        output["length"] = self.length
        output["cutoff"] = self.cutoff
        output["sample_rate"] = self.samplerate
        output["status"] = self.status
        output["system"] = system_summary()
        return output


##
# @brief LPAddNoise
# @details Component to add additional noise to a signal array.
# @par EAGLE_START
# @param category PythonApp
# @param[in] cparam/appclass appclass/dlg_lowpass_components.LPAddNoise/String/readonly/
#     \~English Import direction for application class
# @param[in] cparam/execution_time Execution Time/5/Float/readonly/False//False/
#     \~English Estimated execution time
# @param[in] cparam/num_cpus No. of CPUs/1/Integer/readonly/False//False/
#     \~English Number of cores used
# @param[in] aparam/noise Average noise/0.0/Float/readwrite/
#     \~English The average value of the injected noise signal
# @param[in] aparam/samplerate Sample rate/5000/Integer/readwrite/
#     \~English The sample rate of the signal
# @param[in] aparam/stddiv Standard deviation/1.0/Float/readwrite/
#     \~English The standard deviation of the noise signal
# @param[in] aparam/frequency Noise frequency/1200/Integer/readwrite/
#     \~English The frequency of the noise
# @param[in] aparam/randomseed Random randomseed/42/Integer/readwrite/
#     \~English Random seed of the noise generator
# @param[in] aparam/noisemultiplier Noise multiplier/0.1/Float/readwrite/
#     \~English A gain factor for the injected noise (noisemultiplier).
# @param[in] port/signal Signal/float/
#     \~English Numpy array containing incoming signal (string dump of floats)
# @param[out] port/signal Signal/float/
#     \~English Numpy array containing final signal (purely real (floats))
# @par EAGLE_END
class LPAddNoise(BarrierAppDROP):
    """
    Component to add additional noise to a signal array.
    """

    component_meta = dlg_component(
        "LPAddNoise",
        "Adds noise to a signal generated " "for the low-pass filter example",
        [dlg_batch_input("binary/*", [])],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )

    # default values
    noise = dlg_float_param("noise", 0.0)
    stddiv = dlg_float_param("stddiv", 1.0)
    frequency = dlg_int_param("frequency", 1200)
    samplerate = dlg_int_param("samplerate", 5000)
    randomseed = dlg_int_param("randomseed", 42)
    noisemultiplier = dlg_float_param("noisemultiplier", 0.1)
    signal = np.empty([1])

    def add_noise(self):
        """
        Adds noise at a specified frequency.
        :return: Modified signal
        """
        np.random.seed(self.randomseed)
        samples = self.noisemultiplier * np.random.normal(
            self.noise, self.stddiv, size=len(self.signal)
        )
        for i in range(len(self.signal)):
            samples[i] += np.sin(2 * np.pi * i * self.frequency / self.samplerate)

        out_array = np.empty(self.signal.shape)
        np.add(self.signal, samples, out=out_array)
        self.signal = out_array
        return self.signal

    def get_inputs(self):
        """
        Reads input data into a numpy array.
        :return:
        """
        ins = self.inputs
        if len(ins) != 1:
            raise Exception("Precisely one input required for %r" % self)

        array = np.frombuffer(droputils.allDropContents(ins[0]))
        self.signal = np.frombuffer(array)

    def run(self):
        """
        Called by DALiuGE to start drop execution.
        :return:
        """
        outs = self.outputs
        if len(outs) < 1:
            raise Exception("At least one output required for %r" % self)
        self.get_inputs()
        sig = self.add_noise()
        data = sig.tobytes()
        for output in outs:
            output.len = len(data)
            output.write(data)

    def generate_recompute_data(self):
        return {
            "noise": self.noise,
            "stddiv": self.stddiv,
            "sample_rate": self.samplerate,
            "randomseed": self.randomseed,
            "noisemultiplier": self.noisemultiplier,
            "system": system_summary(),
            "status": self.status,
        }



import numpy as np


def determine_size(length):
    """
    :param length:
    :return: Computes the next largest power of two needed to contain |length| elements
    """
    return int(2 ** np.ceil(np.log2(length))) - 1



"""
dlg_lowpass_components filters
"""
import logging

import numpy as np
import pyfftw
from dlg import droputils
from dlg.common.reproducibility.constants import system_summary
from dlg.drop import BarrierAppDROP
from dlg.meta import (
    dlg_batch_input,
    dlg_batch_output,
    dlg_bool_param,
    dlg_component,
    dlg_streaming_input,
)

from dlg_lowpass_components.utils import determine_size

logger = logging.getLogger(__name__)


##
# @brief LPFilterfftNP
# @details Implements a lowpass filter via fft with numpy
# @par EAGLE_START
# @param category PythonApp
# @param[in] cparam/appclass appclass/dlg_lowpass_components.LPFilterFFTNP/String/readonly/
#     \~English Import direction for application class
# @param[in] cparam/execution_time Execution Time/5/Float/readonly/False//False/
#     \~English Estimated execution time
# @param[in] cparam/num_cpus No. of CPUs/1/Integer/readonly/False//False/
#     \~English Number of cores used
# @param[in] aparam/doubleprecision Double Precision/false/Boolean/readwrite/
#     \~English Whether to use double (true) or float (false) precision.
# @param[in] port/signal Signal/float/
#     \~English Numpy array containing incoming signal (string dump of floats)
# @param[in] port/window Window/float/
#     \~English Numpy array containing the filter window (string dump of floats)
# @param[out] port/signal Signal/Complex/
#     \~English Numpy array containing final signal (complex)
# @par EAGLE_END
class LPFilterFFTNP(BarrierAppDROP):
    """
    Uses numpy to filter a nosiy signal.
    """

    component_meta = dlg_component(
        "LP_filter_np",
        "Filters a signal with " "a provided window using numpy",
        [dlg_batch_input("binary/*", [])],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )

    PRECISIONS = {
        "double": {"float": np.float64, "complex": np.complex128},
        "single": {"float": np.float32, "complex": np.complex64},
    }
    precision = {}
    # default values
    doubleprecision = dlg_bool_param("doubleprecision", True)
    series = []
    output = np.zeros([1])

    def initialize(self, **kwargs):
        super().initialize(**kwargs)
        if self.doubleprecision:
            self.precision = self.PRECISIONS["double"]
        else:
            self.precision = self.PRECISIONS["single"]

    def get_inputs(self):
        """
        Reads input arrays into numpy array
        :return: Sets class series variable.
        """
        ins = self.inputs
        if len(ins) != 2:
            raise Exception("Precisely two input required for %r" % self)

        array = [np.frombuffer(droputils.allDropContents(inp)) for inp in ins]
        self.series = array

    def filter(self):
        """
        Actually performs the filtering
        :return: Numpy array of filtered signal.
        """
        signal = self.series[0]
        window = self.series[1]
        nfft = determine_size(len(signal) + len(window) - 1)
        sig_zero_pad = np.zeros(nfft, dtype=self.precision["float"])
        win_zero_pad = np.zeros(nfft, dtype=self.precision["float"])
        sig_zero_pad[0 : len(signal)] = signal
        win_zero_pad[0 : len(window)] = window
        sig_fft = np.fft.fft(sig_zero_pad)
        win_fft = np.fft.fft(win_zero_pad)
        out_fft = np.multiply(sig_fft, win_fft)
        out = np.fft.ifft(out_fft)
        return out.astype(self.precision["complex"])

    def run(self):
        """
        Called by DALiuGE to start execution
        :return:
        """
        outs = self.outputs
        if len(outs) < 1:
            raise Exception("At least one output required for %r" % self)
        self.get_inputs()
        self.output = self.filter()
        data = self.output.tobytes()
        for output in outs:
            output.len = len(data)
            output.write(data)

    def generate_recompute_data(self):
        return {
            "precision_float": str(self.precision["float"]),
            "precision_complex": str(self.precision["complex"]),
            "system": system_summary(),
            "status": self.status,
        }


##
# @brief LPFilterFFTFFTW
# @details Implements a lowpass filter via fft with FFTW
# @par EAGLE_START
# @param category PythonApp
# @param[in] cparam/appclass appclass/dlg_lowpass_components.LPFilterFFTFFTW/String/readonly/
#     \~English Import direction for application class
# @param[in] cparam/execution_time Execution Time/5/Float/readonly/False//False/
#     \~English Estimated execution time
# @param[in] cparam/num_cpus No. of CPUs/1/Integer/readonly/False//False/
#     \~English Number of cores used
# @param[in] aparam/doubleprecision Double Precision/false/Boolean/readwrite/
#     \~English Whether to use double (true) or float (false) precision.
# @param[in] port/signal Signal/float/
#     \~English Numpy array containing incoming signal (string dump of floats)
# @param[in] port/window Window/float/
#     \~English Numpy array containing the filter window (string dump of floats)
# @param[out] port/signal Signal/Complex/
#     \~English Numpy array containing final signal (complex)
# @par EAGLE_END
class LPFilterFFTFFTW(LPFilterFFTNP):
    """
    Uses fftw to implement a low-pass filter
    """

    component_meta = dlg_component(
        "LP_filter_fftw",
        "Filters a signal with " "a provided window using FFTW",
        [dlg_batch_input("binary/*", [])],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )

    def filter(self):
        """
        Actually performs the filtering
        :return: Filtered signal as numpy array.
        """
        pyfftw.interfaces.cache.disable()
        signal = self.series[0]
        window = self.series[1]
        nfft = determine_size(len(signal) + len(window) - 1)
        sig_zero_pad = pyfftw.empty_aligned(len(signal), dtype=self.precision["float"])
        win_zero_pad = pyfftw.empty_aligned(len(window), dtype=self.precision["float"])
        sig_zero_pad[0 : len(signal)] = signal
        win_zero_pad[0 : len(window)] = window
        sig_fft = pyfftw.interfaces.numpy_fft.fft(sig_zero_pad, n=nfft)
        win_fft = pyfftw.interfaces.numpy_fft.fft(win_zero_pad, n=nfft)
        out_fft = np.multiply(sig_fft, win_fft)
        out = pyfftw.interfaces.numpy_fft.ifft(out_fft, n=nfft)
        return out.astype(self.precision["complex"])


##
# @brief LPFilterFFTCuda
# @details Implements a lowpass filter via fft with cuda
# @par EAGLE_START
# @param category PythonApp
# @param[in] cparam/appclass appclass/dlg_lowpass_components.LPFilterFFTCuda/String/readonly/
#     \~English Import direction for application class
# @param[in] cparam/execution_time Execution Time/5/Float/readonly/False//False/
#     \~English Estimated execution time
# @param[in] cparam/num_cpus No. of CPUs/1/Integer/readonly/False//False/
#     \~English Number of cores used
# @param[in] aparam/doubleprecision Double Precision/true/Boolean/readwrite/
#     \~English Whether to use double (true) or float (false) precision.
# @param[in] port/signal Signal/float/
#     \~English Numpy array containing incoming signal (string dump of floats)
# @param[in] port/window Window/float/
#     \~English Numpy array containing the filter window (string dump of floats)
# @param[out] port/signal Signal/Complex/
#     \~English Numpy array containing final signal (complex)
# @par EAGLE_END
class LPFilterFFTCuda(LPFilterFFTNP):
    """
    Uses pycuda to implement a low-pass filter
    """

    component_meta = dlg_component(
        "LPFilterFFTCuda",
        "Filters a signal with " "a provided window using cuda",
        [dlg_batch_input("binary/*", [])],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )

    def filter(self):
        """
        Actually performs the filtering
        :return:
        """
        import pycuda.gpuarray as gpuarray
        import skcuda.fft as cu_fft
        import skcuda.linalg as linalg
        import pycuda.driver as cuda
        from pycuda.tools import make_default_context

        cuda.init()
        context = make_default_context()
        device = context.get_device()
        signal = self.series[0]
        window = self.series[1]
        linalg.init()
        nfft = determine_size(len(signal) + len(window) - 1)
        # Move data to GPU
        sig_zero_pad = np.zeros(nfft, dtype=self.precision["float"])
        win_zero_pad = np.zeros(nfft, dtype=self.precision["float"])
        sig_gpu = gpuarray.zeros(sig_zero_pad.shape, dtype=self.precision["float"])
        win_gpu = gpuarray.zeros(win_zero_pad.shape, dtype=self.precision["float"])
        sig_zero_pad[0 : len(signal)] = signal
        win_zero_pad[0 : len(window)] = window
        sig_gpu.set(sig_zero_pad)
        win_gpu.set(win_zero_pad)

        # Plan forwards
        sig_fft_gpu = gpuarray.zeros(nfft, dtype=self.precision["complex"])
        win_fft_gpu = gpuarray.zeros(nfft, dtype=self.precision["complex"])
        sig_plan_forward = cu_fft.Plan(
            sig_fft_gpu.shape, self.precision["float"], self.precision["complex"]
        )
        win_plan_forward = cu_fft.Plan(
            win_fft_gpu.shape, self.precision["float"], self.precision["complex"]
        )
        cu_fft.fft(sig_gpu, sig_fft_gpu, sig_plan_forward)
        cu_fft.fft(win_gpu, win_fft_gpu, win_plan_forward)

        # Convolve
        out_fft = linalg.multiply(sig_fft_gpu, win_fft_gpu, overwrite=True)
        linalg.scale(2.0, out_fft)

        # Plan inverse
        out_gpu = gpuarray.zeros_like(out_fft)
        plan_inverse = cu_fft.Plan(
            out_fft.shape, self.precision["complex"], self.precision["complex"]
        )
        cu_fft.ifft(out_fft, out_gpu, plan_inverse, True)
        out_np = np.zeros(len(out_gpu), self.precision["complex"])
        out_gpu.get(out_np)
        context.pop()
        return out_np


##
# @brief LPFilterPointwiseNP
# @details Implements a lowpass filter via pointwise convolution with numpy
# @par EAGLE_START
# @param category PythonApp
# @param[in] cparam/appclass appclass/dlg_lowpass_components.LPFilterPointwiseNP/String/readonly/
#     \~English Import direction for application class
# @param[in] cparam/execution_time Execution Time/5/Float/readonly/False//False/
#     \~English Estimated execution time
# @param[in] cparam/num_cpus No. of CPUs/1/Integer/readonly/False//False/
#     \~English Number of cores used
# @param[in] aparam/doubleprecision Double Precision/false/Boolean/readwrite/
#     \~English Whether to use double (true) or float (false) precision.
# @param[in] port/signal Signal/float/
#     \~English Numpy array containing incoming signal (string dump of floats)
# @param[in] port/window Window/float/
#     \~English Numpy array containing the filter window (string dump of floats)
# @param[out] port/signal Signal/Complex/
#     \~English Numpy array containing final signal (complex)
# @par EAGLE_END
class LPFilterPointwiseNP(LPFilterFFTNP):
    """
    Uses raw numpy to implement a low-pass filter
    """

    component_meta = dlg_component(
        "LPFilterPointwiseNP",
        "Filters a signal with " "a provided window using cuda",
        [dlg_batch_input("binary/*", [])],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )

    def filter(self):
        return np.convolve(self.series[0], self.series[1], mode="full").astype(
            self.precision["complex"]
        )



__package__ = "dlg_lowpass_components"

# The following imports are the binding to the DALiuGE system
from dlg import droputils, utils

# extend the following as required
from .generators import LPSignalGenerator, LPWindowGenerator, LPAddNoise
from .filters import (
    LPFilterFFTNP,
    LPFilterFFTFFTW,
    LPFilterFFTCuda,
    LPFilterPointwiseNP,
)

__all__ = [
    "LPSignalGenerator",
    "LPWindowGenerator",
    "LPAddNoise",
    "LPFilterFFTNP",
    "LPFilterFFTFFTW",
    "LPFilterFFTCuda",
    "LPFilterPointwiseNP",
]



# __main__ is not required for DALiuGE components.
import argparse  # pragma: no cover

from dlg.drop import InMemoryDROP

from dlg_lowpass_components import LPSignalGenerator


def main() -> None:  # pragma: no cover
    """
    The main function executes on commands:
    `python -m dlg_lowpass_components` and `$ dlg_lowpass_components `.

    This is your program's entry point.

    You can change this function to do whatever you want.
    Examples:
        * Run a test suite
        * Run a server
        * Do some other stuff
        * Run a command line application (Click, Typer, ArgParse)
        * List all available tasks
        * Run an application (Flask, FastAPI, Django, etc.)
    """
    parser = argparse.ArgumentParser(
        description="dlg_lowpass_components.",
        epilog="Enjoy the dlg_lowpass_components functionality!",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Optionally adds verbosity",
    )
    args = parser.parse_args()
    print("Executing main function")
    comp = LPSignalGenerator("A", "A")
    memory = InMemoryDROP("b", "b")
    comp.addOutput(memory)
    comp.run()
    print("End of main function")


if __name__ == "__main__":  # pragma: no cover
    main()



import unittest
import time
import numpy as np

from dlg_lowpass_components import (
    LPSignalGenerator,
    LPWindowGenerator,
    LPAddNoise,
    LPFilterFFTFFTW,
    LPFilterFFTCuda,
    LPFilterPointwiseNP,
    LPFilterFFTNP,
)
from dlg.drop import InMemoryDROP
from dlg.droputils import allDropContents, DROPWaiterCtx


def _run_component(component):
    memory = InMemoryDROP("b", "b")
    component.addOutput(memory)
    component.run()
    memory.setCompleted()
    return allDropContents(memory)


class TestLPSignalGenerator(unittest.TestCase):
    def test_default(self):
        generator = LPSignalGenerator("a", "a")
        signal = _run_component(generator)
        self.assertIsNotNone(signal)
        self.assertEqual(generator.length * 8, len(signal))

    def test_changed_length(self):
        new_length = 1024
        generator = LPSignalGenerator("a", "a", length=new_length)
        signal = _run_component(generator)
        self.assertIsNotNone(signal)
        self.assertEqual(new_length, generator.length)
        self.assertEqual(new_length * 8, len(signal))

    def test_added_noise(self):
        noise_params = {"noise": 0.0, "stddiv": 1.0, "frequency": 666, "randomseed": 42}
        noisy_generator = LPSignalGenerator("a", "a", noise=noise_params)
        vanilla_generator = LPSignalGenerator("A", "A")
        noisy_signal = _run_component(noisy_generator)
        vanilla_signal = _run_component(vanilla_generator)
        self.assertIsNotNone(noisy_signal)
        self.assertIsNotNone(vanilla_signal)
        self.assertNotEqual(noisy_signal, vanilla_signal)

    def test_changed_frequency(self):
        new_frequency = {"values": [440, 900, 1100, 1900, 2400]}
        generator = LPSignalGenerator("a", "a", frequencies=new_frequency)
        vanilla_generator = LPSignalGenerator("A", "A")
        signal = _run_component(generator)
        vanilla_signal = _run_component(vanilla_generator)

        self.assertIsNotNone(signal)
        self.assertIsNotNone(vanilla_signal)
        self.assertNotEqual(signal, vanilla_signal)


class TestLPWindowGenerator(unittest.TestCase):
    def test_default(self):
        generator = LPWindowGenerator("a", "a")
        signal = _run_component(generator)
        self.assertIsNotNone(signal)
        self.assertEqual(generator.length * 8, len(signal))

    def test_change_length(self):
        new_length = 512
        generator = LPWindowGenerator("a", "a", length=new_length)
        signal = _run_component(generator)
        self.assertIsNotNone(signal)
        self.assertEqual(new_length, generator.length)
        self.assertEqual(new_length * 8, len(signal))

    def test_change_cutoff(self):
        new_cutoff = 660
        generator = LPWindowGenerator("a", "a", cutoff=new_cutoff)
        vanilla_generator = LPWindowGenerator("A", "A")
        signal = _run_component(generator)
        vanilla_signal = _run_component(vanilla_generator)
        self.assertIsNotNone(signal)
        self.assertIsNotNone(vanilla_signal)
        self.assertNotEqual(signal, vanilla_signal)

    def test_change_srate(self):
        new_srate = 4500
        generator = LPWindowGenerator("a", "a", samplerate=new_srate)
        vanilla_generator = LPWindowGenerator("A", "A")
        signal = _run_component(generator)
        vanilla_signal = _run_component(vanilla_generator)
        self.assertIsNotNone(signal)
        self.assertIsNotNone(vanilla_signal)
        self.assertNotEqual(signal, vanilla_signal)


class TestLPNoiseGenerator(unittest.TestCase):
    def test_compare_to_SignalGen(self):
        noise_params = {"noise": 0.0, "stddiv": 1.0, "frequency": 1200, "randomseed": 42}
        noisy_generator = LPSignalGenerator("a", "a", noise=noise_params)
        noisy_signal = _run_component(noisy_generator)

        vanilla_generator = LPSignalGenerator("A", "A")
        interim_mem = InMemoryDROP("B", "B")
        noisy_adder = LPAddNoise("C", "C")
        final_mem = InMemoryDROP("D", "D")

        vanilla_generator.addOutput(interim_mem)
        noisy_adder.addInput(interim_mem)
        noisy_adder.addOutput(final_mem)

        with DROPWaiterCtx(self, final_mem, 10):
            vanilla_generator.run()
            interim_mem.setCompleted()

        final_signal = allDropContents(final_mem)

        self.assertIsNotNone(noisy_signal)
        self.assertIsNotNone(final_signal)
        self.assertEqual(noisy_signal, final_signal)


def _run_filter(obj, filter):
    generator = LPSignalGenerator("a", "a")
    signal_drop = InMemoryDROP("b", "b")
    window_drop = InMemoryDROP("c", "c")
    in_signal = _run_component(generator)

    generator = LPWindowGenerator("a", "a")
    window = _run_component(generator)
    filtered_signal_drop = InMemoryDROP("d", "d")
    filter.addInput(signal_drop)
    filter.addInput(window_drop)
    filter.addOutput(filtered_signal_drop)

    with DROPWaiterCtx(obj, filtered_signal_drop, 10):
        signal_drop.write(in_signal)
        window_drop.write(window)
        signal_drop.setCompleted()
        window_drop.setCompleted()

    filtered_signal = allDropContents(filtered_signal_drop)
    return filtered_signal, len(in_signal)


class TestLPFilterFFTNP(unittest.TestCase):
    def test_default(self):
        filter = LPFilterFFTNP("a", "a")
        filtered_signal, original_length = _run_filter(self, filter)
        self.assertIsNotNone(filtered_signal)
        self.assertEqual(
            original_length / 8 * 2 - 1,
            len(filtered_signal) / np.dtype(complex).itemsize,
        )

    def test_precision(self):
        float_filter = LPFilterFFTNP("a", "a", doubleprecision=False)
        double_filter = LPFilterFFTNP("a", "a")
        float_signal, _ = _run_filter(self, float_filter)
        double_signal, _ = _run_filter(self, double_filter)
        self.assertNotEqual(float_signal, double_signal)


class TestLPFilterFFTFFTW(unittest.TestCase):
    def test_default(self):
        filter = LPFilterFFTFFTW("a", "a")
        filtered_signal, original_length = _run_filter(self, filter)
        self.assertIsNotNone(filtered_signal)
        self.assertEqual(
            original_length / 8 * 2 - 1,
            len(filtered_signal) / np.dtype(complex).itemsize,
        )

    def test_precision(self):
        float_filter = LPFilterFFTFFTW("a", "a", doubleprecision=False)
        double_filter = LPFilterFFTFFTW("a", "a")
        float_signal, _ = _run_filter(self, float_filter)
        double_signal, _ = _run_filter(self, double_filter)
        self.assertNotEqual(float_signal, double_signal)


class TestLPFilterFFTCuda(unittest.TestCase):
    # WARNING: GCC version 9 or earlier.
    def test_default(self):
        filter = LPFilterFFTCuda("a", "a")
        filtered_signal, original_length = _run_filter(self, filter)
        self.assertIsNotNone(filtered_signal)
        self.assertEqual(
            original_length / 8 * 2 - 1,
            len(filtered_signal) / np.dtype(complex).itemsize,
        )

    def test_precision(self):
        float_filter = LPFilterFFTCuda("a", "a", doubleprecision=False)
        double_filter = LPFilterFFTCuda("a", "a")
        float_signal, _ = _run_filter(self, float_filter)
        double_signal, _ = _run_filter(self, double_filter)
        self.assertNotEqual(float_signal, double_signal)


class TestLPFilterPointwiseNP(unittest.TestCase):
    def test_default(self):
        filter = LPFilterPointwiseNP("a", "a")
        filtered_signal, original_length = _run_filter(self, filter)
        self.assertIsNotNone(filtered_signal)
        self.assertEqual(
            original_length / 8 * 2 - 1,
            len(filtered_signal) / np.dtype(complex).itemsize,
        )

    def test_precision(self):
        float_filter = LPFilterPointwiseNP("a", "a", doubleprecision=False)
        double_filter = LPFilterPointwiseNP("a", "a")
        float_signal, _ = _run_filter(self, float_filter)
        double_signal, _ = _run_filter(self, double_filter)
        self.assertNotEqual(float_signal, double_signal)
